/* Copyright 2005 Garrett Rooney.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <httpd.h>
#include <http_config.h>
#include <http_request.h>
#include <http_log.h>
#include <util_filter.h>

#include <apr_strings.h>
#include <apr_pools.h>

#include "etl_template.h"

extern char **environ;

static int
template_example_file_handler (request_rec *r)
{
  apr_status_t apr_err;
  apr_finfo_t finfo;

  /* first, make sure that we're supposed to handle this, and that we were
   * asked for a template that's actually there. */

  if (! r->handler || strcmp (r->handler, "template-example") != 0)
    return DECLINED;

  r->allowed |= (AP_METHOD_BIT << M_GET);
  if (r->method_number != M_GET)
    return HTTP_METHOD_NOT_ALLOWED;

  apr_err = apr_stat (&finfo, r->filename, APR_FINFO_TYPE, r->pool);
  if (apr_err)
    return HTTP_NOT_FOUND;

  /* ok, now that we're sure we should be handling this, lets do it. */

  {
    apr_hash_t *environment;
    apr_bucket_brigade *out;
    etl_template_t *tmpl;

    /* first parse the template into its compiled form... */

    apr_err = etl_template_parse_file (&tmpl, r->filename, r->pool);
    if (apr_err)
      {
        ap_log_rerror (APLOG_MARK,
                       APLOG_ERR,
                       apr_err,
                       r,
                       "error opening template file: %s",
                       r->filename);

        return HTTP_NOT_FOUND;
      }

    r->content_type = "text/html";

    /* create ourselves an environment full of variables for the templates
     * to play with. */

    environment = apr_hash_make (r->pool);

    {
      etl_template_variable_t *env = etl_template_make_hash (r->pool);
      char **ep;

      for (ep = environ; *ep; ++ep)
        {
          char *tmp, *key, *val;

          key = apr_strtok (apr_pstrdup (r->pool, *ep), "=", &tmp);

          if (! key)
            continue;

          val = apr_strtok (NULL, "=", &tmp);

          apr_hash_set (env->contents.h,
                        key,
                        APR_HASH_KEY_STRING,
                        etl_template_make_str (val, r->pool));
        }

      apr_hash_set (environment, "env", APR_HASH_KEY_STRING, env);
    }

    out = apr_brigade_create (r->pool, r->connection->bucket_alloc);

    /* now execute it, catching the output in a bucket brigade. */

    apr_err = etl_template_execute (tmpl, environment, out, r->pool);
    if (apr_err)
      {
        ap_log_rerror (APLOG_MARK,
                       APLOG_ERR,
                       apr_err,
                       r,
                       "error executing template file: %s",
                       r->filename);

        return HTTP_INTERNAL_SERVER_ERROR;
      }

    /* hand it off to the rest of the output chain. */

    ap_pass_brigade (r->output_filters, out);
  }

  return OK;
}

static void
template_example_register_hooks (apr_pool_t *pool)
{
  ap_hook_handler (template_example_file_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA template_example_module = {
  STANDARD20_MODULE_STUFF,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  template_example_register_hooks
};
