/* Copyright 2005-2006 Garrett Rooney.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdlib.h> /* for abort */

#include "lexer.h"

#include <apr_strings.h>

#define BSIZE          8192
#define MAX_TEXT_BYTES 1024

struct etl_lexer_t {
  apr_bucket_brigade *bb;  /* the brigade we're reading from */
  apr_bucket_brigade *tbb; /* a temp brigade for reading lines */

  /* bookkeeping data for the re2c lexer */
  char *buffer;
  char *cursor;
  char *limit;
  char *marker;
  char *tstart;

  /* bookkeeping data for the stored up text */
  char text_buffer[MAX_TEXT_BYTES];
  int ntext_bytes;
  int saved_bytes;

  int pending_token; /* next token after a text node, saved to return later */

  int eof;          /* have we hit the end of the input? */
  int in_directive; /* are we inside a etl directive? */
};

etl_lexer_t *
etl_lexer_create (apr_bucket_brigade *bb, apr_pool_t *pool)
{
  etl_lexer_t *l = apr_pcalloc (pool, sizeof (*l));

  l->bb = bb;
  l->tbb = apr_brigade_create (pool, bb->bucket_alloc);
  l->buffer = apr_pcalloc (pool, BSIZE);

  return l;
}

static apr_status_t
fill (etl_lexer_t *l)
{
  if (l->cursor == l->limit)
    {
      apr_status_t err = apr_brigade_split_line (l->tbb,
                                                 l->bb,
                                                 APR_BLOCK_READ,
                                                 BSIZE - 1);
      if (err)
        return err;
      else
        {
          apr_size_t len = BSIZE;

          apr_brigade_flatten (l->tbb, l->buffer, &len);

          if (len == 0)
            {
              l->eof = 1;
              l->buffer[0] = '\n';
              l->buffer[1] = '\0';
              l->limit = l->buffer;
            }
          else
            {
              l->limit = l->buffer + len;
            }

          apr_brigade_cleanup (l->tbb);
        }

      l->cursor = l->marker = l->tstart = l->buffer;
    }

  return APR_SUCCESS;
}

/* Defines needed for re2c. */
#define  YYCTYPE   char
#define  YYCURSOR  l->cursor
#define  YYLIMIT   l->limit
#define  YYMARKER  l->marker
#define  YYFILL(n) { apr_status_t rv = fill(l); if (rv) return rv; }

/* A little shorthand... */
#define RET(i) do {                                   \
                 if (l->ntext_bytes)                  \
                   {                                  \
                     l->pending_token = i;            \
                     *t = ETL_TOK_TEXT;               \
                     l->saved_bytes = l->ntext_bytes; \
                     l->ntext_bytes = 0;              \
                   }                                  \
                 else                                 \
                   {                                  \
                     *t = i;                          \
                   }                                  \
                 return APR_SUCCESS;                  \
               } while (0)

apr_status_t
etl_lexer_scan (etl_token_t *t, etl_lexer_t *l, apr_pool_t *pool)
{
  if (l->pending_token)
    {
      *t = l->pending_token;

      l->pending_token = 0;

      return APR_SUCCESS;
    }

start:
  l->tstart = l->cursor;

  if (! l->in_directive)
    {
      /*!re2c

        identifier    = [a-zA-Z\.,\200-\377]*;
        any           = [\000-\377];
        newline       = "\n";
        whitespace    = [ \t\r];

        "[%"          {
                        l->in_directive = 1;
                        RET(ETL_TOK_START_DIR);
                      }

        newline       {
                        if (l->eof)
                          RET(ETL_TOK_EOI);
                        else
                          goto text;
                      }

        any           {
                      text:
                        l->text_buffer[l->ntext_bytes] = *(YYCURSOR - 1);
                        l->ntext_bytes++;

                        if (l->ntext_bytes == MAX_TEXT_BYTES)
                          {
                            *t = ETL_TOK_TEXT;
                            l->saved_bytes = l->ntext_bytes;
                            l->ntext_bytes = 0;
                            return APR_SUCCESS;
                          }
                        else
                          goto start;
                      }
       */
    }
  else
    {
      /*!re2c

        "%]"          {
                        l->in_directive = 0;
                        RET(ETL_TOK_END_DIR);
                      }
        "if"          { RET(ETL_TOK_IF); }
        "in"          { RET(ETL_TOK_IN); }
        "unless"      { RET(ETL_TOK_UNLESS); }
        "end"         { RET(ETL_TOK_END); }
        "for"         { RET(ETL_TOK_FOR); }
        "else"        { RET(ETL_TOK_ELSE); }
        "print"       { RET(ETL_TOK_PRINT); }
        "|"           { RET(ETL_TOK_PIPE); }
        whitespace    { goto start; }
        identifier    { RET(ETL_TOK_VAR); }
        newline       {
                        if (l->eof)
                          RET(ETL_TOK_EOI);
                        else
                          goto start;
                      }
        any           {
                        /* We don't understand it, so... */
                        return APR_EINVAL;
                      }
      */
    }

  abort();
}

char *
etl_lexer_text (etl_lexer_t *lexer, apr_pool_t *pool)
{
  if (lexer->saved_bytes)
    {
      char *str = apr_pcalloc (pool, lexer->saved_bytes + 1);

      memcpy (str, lexer->text_buffer, lexer->saved_bytes);

      lexer->saved_bytes = 0;

      return str;
    }
  else
    {
      apr_size_t len = lexer->cursor - lexer->tstart;

      char *str = apr_pcalloc (pool, len + 1);

      memcpy (str, lexer->tstart, len);

      return str;
    }
}
